
$x = 7;
$y = 3 * $x + 4;
$z = 3 * ($x + 2);

$x = $x + 1 * $y + $z;

if ($x > 10) {
    print "The value of \$x is $x\n";
}

print "The value of \$y is $y\n";

if ($z > 100) {
    print "The value of \$z is $z\n";
}

